/*
 *  OpenPrinting Vector Printer Driver for Canon Pixus850i
 *  Copyright (c) 2004 E&D Inc.
 *  All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * NOTE:
 *  - As a special exception, this program is permissible to link with the
 *    libraries released as the binary modules.
 *  - If you write modifications of your own for these programs, it is your
 *    choice whether to permit this exception to apply to your modifications.
 *    If you do not wish that, delete this exception.
*/

/*
 * Common Functions : Data Management
 */

/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"

/*==================================================================
  Printer Infomation
===================================================================*/
pd_PrinterInfoList	*pd_gPrinterList = NULL;

/*******************************************************************
  for Data Convert
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_DtoI
 *
 * ---------------------------------------------------------------- */
int	pd_DtoI( double d )
{
	int		s, i;
	s = 0;
	if( d < 0 ) {	s = 1;	d *= -1;	}
	i = (int)( d + 0.5 );
	if( s )		i *= -1;
	return( i );
}

/*******************************************************************
  for Data Management
********************************************************************/

/* ----------------------------------------------------------------
 *
 *	pd_ClearGraphicsState
 *
 * ---------------------------------------------------------------- */
void	pd_ClearGraphicsState( pd_GraphicsState *dst )
{
	memset( dst, 0, sizeof( pd_GraphicsState ) );
}
/* ----------------------------------------------------------------
 *
 *	pd_CopyGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_CopyGraphicsState( pd_GraphicsState *dst, pd_GraphicsState *src )
{
	pd_ClearGraphicsState( dst );
	memcpy( dst, src, sizeof( pd_GraphicsState ) );
	return( TRUE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pd_GetPrinterInfo
 *
 * ---------------------------------------------------------------- */
pd_PrinterInfo	*pd_GetPrinterInfo( int printerContext )
{
	pd_PrinterInfoList	*pls;

	for( pls = pd_gPrinterList; pls; pls = pls->next ) {
		if( pls->printer.printerContext == printerContext )	break;
	}
	if( !pls ) {
		errorno = pdBADCONTEXT;
		return( NULL );
	}
	return( &pls->printer );
}

/* ----------------------------------------------------------------
 *
 *	pd_NewPrinterInfo
 *
 * ---------------------------------------------------------------- */
pd_PrinterInfo	*pd_NewPrinterInfo( int fd )
{
	pd_PrinterInfoList	**plp, *pls;
	pd_PrinterInfo		*pri;
	int					pct;

	pct = 0;
	for( plp = &pd_gPrinterList; *plp; plp = &(*plp)->next ) {
		if( (*plp)->printer.outputFD == fd ) {
			errorno = pdBADREQUEST;
			return( NULL );
		}
		if( (*plp)->printer.printerContext > pct )
			pct = (*plp)->printer.printerContext;
	}
	pct++;

	pls = (pd_PrinterInfoList *)malloc( sizeof( pd_PrinterInfoList ) );
	if( !pls ) {
		errorno = pdFATALERROR;
		return( NULL );
	}
	memset( pls, 0, sizeof( pd_PrinterInfoList ) );
	pri = &pls->printer;
	pri->printerContext = pct;
	pri->outputFD = fd;
	*plp = pls;
	return( &pls->printer );
}

/* ----------------------------------------------------------------
 *
 *	pd_DeletePrinterInfo
 *
 * ---------------------------------------------------------------- */
bool	pd_DeletePrinterInfo( int printerContext )
{
	pd_PrinterInfoList		**plp, *pls;
	pd_PrinterInfo			*pri;
	pd_GraphicsStateList	*gls, *gln;

	for( plp = &pd_gPrinterList; *plp; plp = &(*plp)->next ) {
		if( (*plp)->printer.printerContext == printerContext )	break;
	}
	pls = *plp;
	if( !pls ) {
		errorno = pdBADCONTEXT;
		return( FALSE );
	}
	*plp = pls->next;
	pri = &pls->printer;

	// Clear GraphicsStateList
	for( gls = pri->gstateList.next; gls; gls = gln ) {
		gln = gls->next;
		pd_ClearGraphicsState( &gls->gstate );
		free( gls );
	}
	pri->gstateList.next = NULL;
	pd_ClearGraphicsState( &pri->gstateList.gstate );

	free( pls );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pd_GetGraphicsState
 *
 * ---------------------------------------------------------------- */
pd_GraphicsState	*pd_GetGraphicsState( int printerContext )
{
	pd_PrinterInfo	*pri;

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( NULL );
	return( &pri->gstateList.gstate );
}

/* ----------------------------------------------------------------
 *
 *	pd_InitGraphicsState
 *
 * ---------------------------------------------------------------- */

void	pd_InitGraphicsState( pd_GraphicsState *gs )
{
	gs->colorSpace = pdcspaceStandardRGB;
}

/* ----------------------------------------------------------------
 *
 *	pd_PushGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_PushGraphicsState( pd_PrinterInfo *pri )
{
	pd_GraphicsStateList	*gls;

	gls = (pd_GraphicsStateList *)malloc( sizeof( pd_GraphicsStateList ) );
	if( !gls ) {
		errorno = pdFATALERROR;
		return( FALSE );
	}
	memset( gls, 0, sizeof( pd_GraphicsStateList ) );
	if( !pd_CopyGraphicsState( &gls->gstate, &pri->gstateList.gstate ) ) {
		free( gls );
		return( FALSE );
	}
	gls->next = pri->gstateList.next;
	pri->gstateList.next = gls;
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pd_PopGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_PopGraphicsState( pd_PrinterInfo *pri )
{
	pd_GraphicsStateList	*gls;

	gls = pri->gstateList.next;
	if( !gls ) {
		errorno = pdBADREQUEST;
		return( FALSE );
	}
	if( !pd_CopyGraphicsState( &pri->gstateList.gstate, &gls->gstate ) ) {
		return( FALSE );
	}
	pri->gstateList.next = gls->next;
	pd_ClearGraphicsState( &gls->gstate );
	free( gls );
	return( TRUE );
}

