/* Copyright (c) 2004 E-D Inc.  All rights reserved.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/
/*******************************************************************

  OpenPrinting Vector Printer Driver API : for Canon

 ==================================================================

  Common Functions : Data Management

********************************************************************/
/* system include files */
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>

/* local include files */
#include "pdapi.h"
#include "pd_print.h"

/*==================================================================
  Printer Infomation
===================================================================*/
pd_PrinterInfoList	*pd_gPrinterList = NULL;

/*******************************************************************
  for Debug
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_DebugPrint
 *
 * ---------------------------------------------------------------- */
void	pd_DebugPrint( const char *fmt, ... )
{
#ifdef DEBUG
	va_list	ap;

	va_start( ap, fmt );
	vfprintf( stderr, fmt, ap );
	va_end( ap );
	fflush( stderr );
#endif
	return;
}

/*******************************************************************
  for Data Convert
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_DtoI
 *
 * ---------------------------------------------------------------- */
int	pd_DtoI( double d )
{
	int		s, i;
	s = 0;
	if( d < 0 ) {	s = 1;	d *= -1;	}
	i = (int)( d + 0.5 );
	if( s )		i *= -1;
	return( i );
}

/* ----------------------------------------------------------------
 *
 *	pd_Dx_dLcd / pd_Dy_dLcd
 *
 * ---------------------------------------------------------------- */
int	pd_Dx_dLcd( double dx, double dy, pdCTM *ctm )
{
	return( pd_DtoI( dx * ctm->a + dy * ctm->c + ctm->e ) );
}
int	pd_Dy_dLcd( double dx, double dy, pdCTM *ctm )
{
	return( pd_DtoI( dx * ctm->b + dy * ctm->d + ctm->f ) );
}
/* ----------------------------------------------------------------
 *
 *	pd_Dx_Lcd / pd_Dy_Lcd
 *
 * ---------------------------------------------------------------- */
int	pd_Dx_Lcd( pdFix fx, pdFix fy, pdCTM *ctm )
{	return( pd_Dx_dLcd( Fixtof( fx ), Fixtof( fy ), ctm ) );	}
int	pd_Dy_Lcd( pdFix fx, pdFix fy, pdCTM *ctm )
{	return( pd_Dy_dLcd( Fixtof( fx ), Fixtof( fy ), ctm ) );	}

/*******************************************************************
  for Data Management
********************************************************************/
/* ----------------------------------------------------------------
 *
 *	pd_CreateBrushData
 *
 * ---------------------------------------------------------------- */
pdBrushData	*pd_CreateBrushData( pdBrushData *src )
{
	pdBrushData	*dst;
	int			siz;

	siz = sizeof( pdBrushData ) + src->pitch * src->height;
	dst = (pdBrushData *)malloc( siz );
	if( !dst ) {
		errorno = pdFATALERROR;
		return( NULL );
	}
	memcpy( dst, src, siz );
	return( dst );
}
/* ----------------------------------------------------------------
 *
 *	pd_EraseBrushData
 *
 * ---------------------------------------------------------------- */
void	pd_EraseBrushData( pdBrushData **obj )
{
	if( *obj ) {	free( *obj );	*obj = NULL;	}
}

/* ----------------------------------------------------------------
 *
 *	pd_CopyBrush
 *
 * ---------------------------------------------------------------- */
bool	pd_CopyBrush( pdBrush *dst, pdBrush *src )
{
	pdBrushData	*brDat = NULL;

	if( src->pbrush ) {
		if( !( brDat = pd_CreateBrushData( src->pbrush ) ) )	return( FALSE );
	}
	pd_EraseBrushData( &dst->pbrush );
	memcpy( dst, src, sizeof( pdBrush ) );
	dst->pbrush = brDat;
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pd_ClearGraphicsState
 *
 * ---------------------------------------------------------------- */
void	pd_ClearGraphicsState( pd_GraphicsState *dst )
{
	pd_EraseBrushData( &dst->strokeBrush.pbrush );
	pd_EraseBrushData( &dst->fillBrush.pbrush );
	pd_EraseBrushData( &dst->bgBrush.pbrush );
	memset( dst, 0, sizeof( pd_GraphicsState ) );
}
/* ----------------------------------------------------------------
 *
 *	pd_CopyGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_CopyGraphicsState( pd_GraphicsState *dst, pd_GraphicsState *src )
{
	pdBrushData	*strokeBrDat = NULL;
	pdBrushData	*fillBrDat = NULL;
	pdBrushData	*bgBrDat = NULL;

	if( src->strokeBrush.pbrush ) {
		if( !( strokeBrDat = pd_CreateBrushData( src->strokeBrush.pbrush ) ) ) {
			return( FALSE );
		}
	}
	if( src->fillBrush.pbrush ) {
		if( !( fillBrDat = pd_CreateBrushData( src->fillBrush.pbrush ) ) ) {
			pd_EraseBrushData( &strokeBrDat );
			return( FALSE );
		}
	}
	if( src->bgBrush.pbrush ) {
		if( !( bgBrDat = pd_CreateBrushData( src->bgBrush.pbrush ) ) ) {
			pd_EraseBrushData( &fillBrDat );
			pd_EraseBrushData( &strokeBrDat );
			return( FALSE );
		}
	}
	pd_ClearGraphicsState( dst );
	memcpy( dst, src, sizeof( pd_GraphicsState ) );
	dst->strokeBrush.pbrush = strokeBrDat;
	dst->fillBrush.pbrush = fillBrDat;
	dst->bgBrush.pbrush = bgBrDat;
	return( TRUE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pd_GetPrinterInfo
 *
 * ---------------------------------------------------------------- */
pd_PrinterInfo	*pd_GetPrinterInfo( int printerContext )
{
	pd_PrinterInfoList	*pls;

	for( pls = pd_gPrinterList; pls; pls = pls->next ) {
		if( pls->printer.printerContext == printerContext )	break;
	}
	if( !pls ) {
		errorno = pdBADCONTEXT;
		return( NULL );
	}
	return( &pls->printer );
}

/* ----------------------------------------------------------------
 *
 *	pd_NewPrinterInfo
 *
 * ---------------------------------------------------------------- */
pd_PrinterInfo	*pd_NewPrinterInfo( int fd )
{
	pd_PrinterInfoList	**plp, *pls;
	pd_PrinterInfo		*pri;
	int					pct;

	pct = 0;
	for( plp = &pd_gPrinterList; *plp; plp = &(*plp)->next ) {
		if( (*plp)->printer.outputFD == fd ) {
			errorno = pdBADREQUEST;
			return( NULL );
		}
		if( (*plp)->printer.printerContext > pct )
			pct = (*plp)->printer.printerContext;
	}
	pct++;

	pls = (pd_PrinterInfoList *)malloc( sizeof( pd_PrinterInfoList ) );
	if( !pls ) {
		errorno = pdFATALERROR;
		return( NULL );
	}
	memset( pls, 0, sizeof( pd_PrinterInfoList ) );
	pri = &pls->printer;
	pri->printerContext = pct;
	pri->outputFD = fd;
	*plp = pls;
	return( &pls->printer );
}

/* ----------------------------------------------------------------
 *
 *	pd_DeletePrinterInfo
 *
 * ---------------------------------------------------------------- */
bool	pd_DeletePrinterInfo( int printerContext )
{
	pd_PrinterInfoList		**plp, *pls;
	pd_PrinterInfo			*pri;
	pd_GraphicsStateList	*gls, *gln;

	for( plp = &pd_gPrinterList; *plp; plp = &(*plp)->next ) {
		if( (*plp)->printer.printerContext == printerContext )	break;
	}
	pls = *plp;
	if( !pls ) {
		errorno = pdBADCONTEXT;
		return( FALSE );
	}
	*plp = pls->next;
	pri = &pls->printer;

	// Clear GraphicsStateList
	for( gls = pri->gstateList.next; gls; gls = gln ) {
		gln = gls->next;
		pd_ClearGraphicsState( &gls->gstate );
		free( gls );
	}
	pri->gstateList.next = NULL;
	pd_ClearGraphicsState( &pri->gstateList.gstate );

	free( pls );
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pd_GetGraphicsState
 *
 * ---------------------------------------------------------------- */
pd_GraphicsState	*pd_GetGraphicsState( int printerContext )
{
	pd_PrinterInfo	*pri;

	pri = pd_GetPrinterInfo( printerContext );
	if( !pri )	return( NULL );
	return( &pri->gstateList.gstate );
}

/* ----------------------------------------------------------------
 *
 *	pd_InitCTM
 *
 * ---------------------------------------------------------------- */
void	pd_InitCTM( pdCTM *ctm )
{
	ctm->a = 1;	ctm->b = 0;
	ctm->c = 0;	ctm->d = 1;
	ctm->e = 0;	ctm->f = 0;
}
/* ----------------------------------------------------------------
 *
 *	pd_InitBrush
 *
 * ---------------------------------------------------------------- */
void	pd_InitBrush( pdBrush *brs, int r, int g, int b )
{
	brs->colorSpace = pdcspaceStandardRGB;
	brs->color[0] = b;
	brs->color[1] = g;
	brs->color[2] = r;
	brs->color[3] = 0;
	brs->xorg = 0;
	brs->yorg = 0;
	brs->pbrush = NULL;
}
/* ----------------------------------------------------------------
 *
 *	pd_InitGraphicsState
 *
 * ---------------------------------------------------------------- */
void	pd_InitGraphicsState( pd_GraphicsState *gs )
{
	pd_InitCTM( &gs->ctm );
	gs->colorSpace = pdcspaceStandardRGB;
	// gs->rop = 0;
	gs->fillMode = pdfillModeEvenOdd;
	// gs->alpha = 0.0;
	itoFix( 1, gs->lineWidth );
	gs->dashNum = 0;
	itoFix( 0, gs->dashOffset );
	gs->lineStyle = pdlineStyleSolid;
	gs->lineCap = pdlineCapButt;
	gs->lineJoin = pdlineJoinMiter;
	itoFix( 0, gs->miterLimit );
	gs->paintMode = pdpaintModeOpaque;
	pd_InitBrush( &gs->strokeBrush, 0, 0, 0 );
	pd_InitBrush( &gs->fillBrush, 0, 0, 0 );
	// pd_InitBrush( &gs->bgBrush, 0, 0, 0 );
	gs->pathEntry = FALSE;
	gs->clipPathEntry = FALSE;
}

/* ----------------------------------------------------------------
 *
 *	pd_PushGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_PushGraphicsState( pd_PrinterInfo *pri )
{
	pd_GraphicsStateList	*gls;

	gls = (pd_GraphicsStateList *)malloc( sizeof( pd_GraphicsStateList ) );
	if( !gls ) {
		errorno = pdFATALERROR;
		return( FALSE );
	}
	memset( gls, 0, sizeof( pd_GraphicsStateList ) );
	if( !pd_CopyGraphicsState( &gls->gstate, &pri->gstateList.gstate ) ) {
		free( gls );
		return( FALSE );
	}
	gls->next = pri->gstateList.next;
	pri->gstateList.next = gls;
	return( TRUE );
}

/* ----------------------------------------------------------------
 *
 *	pd_PopGraphicsState
 *
 * ---------------------------------------------------------------- */
bool	pd_PopGraphicsState( pd_PrinterInfo *pri )
{
	pd_GraphicsStateList	*gls;

	gls = pri->gstateList.next;
	if( !gls ) {
		errorno = pdBADREQUEST;
		return( FALSE );
	}
	if( !pd_CopyGraphicsState( &pri->gstateList.gstate, &gls->gstate ) ) {
		return( FALSE );
	}
	pri->gstateList.next = gls->next;
	pd_ClearGraphicsState( &gls->gstate );
	free( gls );
	return( TRUE );
}

/*==================================================================*/
/* ----------------------------------------------------------------
 *
 *	pd_GetNewPathID
 *
 * ---------------------------------------------------------------- */
int	pd_GetNewPathID( pd_PrinterInfo *pri )
{
	pd_GraphicsStateList	*gls;
	char	idf[16];
	int		n;

	memset( idf, 0, 16 );
	if( pri->gstateList.gstate.clipPathEntry ) {
		idf[ pri->gstateList.gstate.clipPathID ] = 1;
	}
	for( gls = pri->gstateList.next; gls; gls = gls->next ) {
		if( gls->gstate.pathEntry ) {
			idf[ gls->gstate.pathID ] = 1;
		}
		if( gls->gstate.clipPathEntry ) {
			idf[ gls->gstate.clipPathID ] = 1;
		}
	}
	for( n = 0; n < 16; n++ ) {
		if( !idf[n] )	break;
	}
	if( n >= 16 )	return( -1 );	// Error
	return( n );
}
